<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Entity;

use Drupal\Core\Entity\EditorialContentEntityBase;
use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityPublishedTrait;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\Plugin\DataType\BooleanData;
use Drupal\Core\TypedData\PrimitiveInterface;
use Drupal\user\EntityOwnerTrait;
use Drupal\user\UserInterface;
use Drupal\workflows\Entity\Workflow;
use Drupal\workflows\TransitionInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workflows\WorkflowTypeInterface;
use Drupal\workspace_approval\Plugin\QueueWorker\RequesterNotificationQueueWorker;
use Drupal\workspace_approval\WorkspaceApprovalNotifications;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;
use Drupal\workspaces\WorkspaceInterface;

/**
 * Defines the workspace approval request entity class.
 *
 * @ContentEntityType(
 *   id = "workspace_approval_request",
 *   label = @Translation("Workspace approval request"),
 *   label_collection = @Translation("Workspace approval requests"),
 *   label_singular = @Translation("workspace approval request"),
 *   label_plural = @Translation("workspace approval requests"),
 *   label_count = @PluralTranslation(
 *     singular = "@count workspace approval requests",
 *     plural = "@count workspace approval requests",
 *   ),
 *   bundle_label = @Translation("Workspace approval request type"),
 *   handlers = {
 *     "list_builder" = "Drupal\workspace_approval\WorkspaceApprovalRequestListBuilder",
 *     "list_builder_reviewer" = "Drupal\workspace_approval\WorkspaceApprovalRequestReviewerListBuilder",
 *     "views_data" = "Drupal\views\EntityViewsData",
 *     "form" = {
 *       "add" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestForm",
 *       "edit" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestForm",
 *       "approve" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestApproveForm",
 *       "reject" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestRejectForm",
 *       "dismiss" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestDismissForm",
 *       "delete" = "Drupal\Core\Entity\ContentEntityDeleteForm",
 *       "delete-multiple-confirm" = "Drupal\Core\Entity\Form\DeleteMultipleForm",
 *     },
 *     "route_provider" = {
 *       "html" = "Drupal\workspace_approval\Routing\WorkspaceApprovalRequestHtmlRouteProvider",
 *     },
 *     "access" = "Drupal\workspace_approval\WorkspaceApprovalRequestAccessControlHandler",
 *     "workspace" = "Drupal\workspaces\Entity\Handler\IgnoredWorkspaceHandler",
 *     "storage" = "Drupal\workspace_approval\WorkspaceApprovalRequestStorage",
 *   },
 *   base_table = "workspace_approval_request",
 *   data_table = "workspace_approval_request_field_data",
 *   revision_table = "workspace_approval_request_revision",
 *   revision_data_table = "workspace_approval_request_field_revision",
 *   admin_permission = "administer workspace_approval_request types",
 *   entity_keys = {
 *     "id" = "id",
 *     "revision" = "vid",
 *     "bundle" = "bundle",
 *     "uuid" = "uuid",
 *     "owner" = "uid",
 *     "published" = "active",
 *   },
 *   revision_metadata_keys = {
 *     "revision_user" = "revision_user",
 *     "revision_created" = "revision_created",
 *     "revision_log_message" = "revision_log_message",
 *   },
 *   show_revision_ui = false,
 *   token_type = "workspace_approval_request",
 *   links = {
 *     "add-form" = "/admin/config/workflow/workspaces/manage/{workspace}/workspace_approval/request/{workspace_approval_request_type}",
 *     "add-page" = "/admin/config/workflow/workspaces/manage/{workspace}/workspace_approval/request",
 *     "canonical" = "/admin/content/workspace_approval/{workspace_approval_request}",
 *     "dismiss-form" = "/admin/content/workspace_approval/{workspace_approval_request}/dismiss",
 *     "approve-form" = "/admin/content/workspace_approval/{workspace_approval_request}/approve",
 *     "reject-form" = "/admin/content/workspace_approval/{workspace_approval_request}/reject",
 *     "delete-form" = "/admin/content/workspace_approval/{workspace_approval_request}/delete",
 *   },
 *   bundle_entity_type = "workspace_approval_request_type",
 * )
 */
final class WorkspaceApprovalRequest extends EditorialContentEntityBase implements WorkspaceApprovalRequestInterface {

  use EntityChangedTrait;
  use EntityOwnerTrait;
  use EntityPublishedTrait;

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage): void {
    parent::preSave($storage);
    if (\is_null($this->getOwnerId())) {
      // If no owner has been set explicitly, make the anonymous user the owner.
      $this->setOwnerId(0);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function label(): TranslatableMarkup {
    $bundle_info = $this->entityTypeBundleInfo()->getBundleInfo($this->getEntityTypeId());
    $bundle_label = $bundle_info[$this->bundle()]['label'];
    return new TranslatableMarkup('@bundle_label from @reviewer', [
      '@bundle_label' => $bundle_label,
      '@reviewer' => $this->getReviewer()?->getDisplayName(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getReviewer(): ?UserInterface {
    $reviewer = $this->get('reviewer')->entity;
    \assert($reviewer instanceof UserInterface);
    return $reviewer;
  }

  /**
   * {@inheritdoc}
   */
  public function getReviewerId(): int {
    $reviewer = $this->get('reviewer')->first();
    \assert($reviewer instanceof EntityReferenceItem);
    $targetId = $reviewer->get('target_id');
    \assert($targetId instanceof PrimitiveInterface);
    return $targetId->getCastedValue();
  }

  /**
   * {@inheritdoc}
   */
  public function getSubject(): ?WorkspaceInterface {
    return $this->get('subject')->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus(): ?bool {
    $status = $this->get('approved');
    \assert($status instanceof FieldItemListInterface);
    if ($status->isEmpty()) {
      return NULL;
    }
    $approved = $status->first()?->get('value');
    \assert($approved instanceof BooleanData);
    return $approved->getCastedValue();
  }

  /**
   * {@inheritdoc}
   */
  public function isApproved(): bool {
    return $this->getStatus() === TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function isRejected(): bool {
    return $this->getStatus() === FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function isPending(): bool {
    return $this->getStatus() === NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getMessageTemplate(WorkspaceApprovalNotifications $type): ?string {
    return $this->getApprovalType()->messageTemplate($type);
  }

  /**
   * {@inheritdoc}
   */
  public function getApprovalType(): WorkspaceApprovalRequestType {
    $request_type = $this->entityTypeManager()->getStorage('workspace_approval_request_type')->load($this->bundle());
    \assert($request_type instanceof WorkspaceApprovalRequestType);
    return $request_type;
  }

  /**
   * {@inheritdoc}
   */
  public function isApprovalRequired(?WorkflowInterface $workflow = NULL, ?TransitionInterface $transition = NULL): bool {
    if ($this->isApproved()) {
      // An approved request will never block a transition.
      return FALSE;
    }
    $workspace = $this->getSubject();
    if (empty($workspace)) {
      // Nothing to guard.
      return FALSE;
    }
    $approvalType = $this->getApprovalType();
    $guardedWorkflowId = $approvalType->workflowId();
    $guardedTransitionId = $approvalType->transitionId();
    \assert(\is_string($guardedWorkflowId) && \is_string($guardedTransitionId));
    if (empty($workflow)) {
      // Check the configured workflow if no workflow was given as an argument.
      $workflow = Workflow::load($guardedWorkflowId);
      \assert($workflow instanceof WorkflowInterface);
    }
    elseif ($workflow->id() !== $guardedWorkflowId) {
      // A workflow was given, but it doesn't match the guarded workflow.
      return FALSE;
    }
    $workflowPlugin = $workflow->getTypePlugin();
    \assert($workflowPlugin instanceof WorkflowTypeInterface);
    if (empty($transition)) {
      // If no transition was given as an argument, then check the transition
      // guarded by this approval type.
      $transition = $workflowPlugin->getTransition($guardedTransitionId);
    }
    else {
      // Ensure that the transition given belongs to the workflow to check.
      if (!$workflowPlugin->hasTransition($transition->id())) {
        throw new \InvalidArgumentException(\sprintf('The %s workflow does not have a %s transition', $guardedWorkflowId, $transition->id()));
      }
      if ($transition->id() !== $guardedTransitionId) {
        // The transition is not guarded by this approval type.
        return FALSE;
      }
    }
    // At this point, the workflow and transition to check have been validated
    // an the only remaining check is whether the workspace has the guarded
    // transition available in its current state.
    $workspaceState = \entity_workflow_get_entity_state($workspace, $guardedWorkflowId);
    \assert(!empty($workspaceState));
    $availableTransitions = $workflowPlugin->getTransitionsForState($workspaceState);
    // This approval type is only required when its guarded transition is
    // currently available.
    return \array_key_exists($transition->id(), $availableTransitions);
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type): array {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    $fields += self::ownerBaseFieldDefinitions($entity_type);
    $fields[$entity_type->getKey('owner')]
      ->setLabel(\t('Requester'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'author',
        'weight' => 15,
      ]);

    $fields += self::publishedBaseFieldDefinitions($entity_type);
    $fields[$entity_type->getKey('published')]
      ->setLabel(\t('Active'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE)
      ->setRevisionable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'entity_reference_label',
        'weight' => 15,
        'settings' => ['format' => 'default'],
      ])
      ->setDefaultValue(TRUE)
      ->setSetting('on_label', 'Active')
      ->setSetting('off_label', 'Dismissed');

    $fields['approved'] = BaseFieldDefinition::create('boolean')
      ->setLabel(\t('Approval status'))
      ->setDisplayConfigurable('form', FALSE)
      ->setRevisionable(TRUE)
      ->setSetting('on_label', 'Approved')
      ->setSetting('off_label', 'Rejected')
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'entity_reference_label',
        'weight' => 15,
        'settings' => ['format' => 'default'],
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['subject'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(\t('Review of'))
      ->setSetting('target_type', 'workspace')
      ->setDefaultValueCallback(self::class . '::getDefaultSubject')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'entity_reference_label',
        'weight' => 15,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['reviewer'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(\t('Reviewer'))
      ->setDescription('The user who will approve or reject the content')
      ->setSetting('target_type', 'user')
      ->setSetting('handler', 'workspace_approval:reviewer')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('form', [
        'type' => 'select2_entity_reference',
        'weight' => 15,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'author',
        'weight' => 15,
      ]);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(\t('Requested on'))
      ->setDescription(\t('The time that the workspace approval request was created.'))
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'timestamp',
        'weight' => 20,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(\t('Updated'))
      ->setDescription(\t('The time that the workspace approval request was last edited.'));

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(EntityStorageInterface $storage, $update = TRUE): void {
    if ($update) {
      \Drupal::queue(RequesterNotificationQueueWorker::ID)->createItem([
        'workspace_approval_request_id' => $this->id(),
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function urlRouteParameters($rel): array {
    $parameters = parent::urlRouteParameters($rel);
    if (\is_null($this->getSubject())) {
      // Nothing to do.
      return $parameters;
    }
    return $parameters + [
      'workspace' => $this->getSubject()->id(),
    ];
  }

  /**
   * Assigns this approval request to a particular workspace.
   *
   * @return string|null
   *   The workspace ID or NULL if a workspace cannot be determined.
   */
  public static function getDefaultSubject(): ?string {
    return \Drupal::routeMatch()->getRawParameter('workspace');
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTagsToInvalidate(): array {
    // Add the subject entity's tags so that when a request is approved or
    // rejected, the entity's workflow transition's local actions will also be
    // invalidated. This ensures that the available transition buttons are kept
    // up to date as requests are updated.
    return [
      ...parent::getCacheTagsToInvalidate(),
      ...$this->getSubject()?->getCacheTagsToInvalidate() ?? [],
    ];
  }

}
