<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Entity\ContentEntityConfirmFormBase;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\entity_workflow\EntityWorkflowTransition;
use Drupal\workflows\Entity\Workflow;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;

/**
 * Base class for forms confirming an operation on a workflow approval request.
 */
abstract class WorkspaceApprovalRequestConfirmForm extends ContentEntityConfirmFormBase {

  /**
   * The guarded workflow transition.
   */
  protected ?WorkflowInterface $workflow = NULL;

  /**
   * The guarded workflow transition.
   */
  protected ?EntityWorkflowTransition $transition = NULL;

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityRepositoryInterface $entity_repository, EntityTypeBundleInfoInterface $entity_type_bundle_info, TimeInterface $time) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    \assert(!empty($this->operation));
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $entity = $this->entity;
    \assert($entity instanceof WorkspaceApprovalRequestInterface);
    $this->workflow = Workflow::load('workspace');
    \assert($this->workflow instanceof WorkflowInterface);
    $approvalRequestType = WorkspaceApprovalRequestType::load($entity->bundle());
    \assert($approvalRequestType instanceof WorkspaceApprovalRequestType);
    $transition = $approvalRequestType->transitionId();
    \assert(\is_string($transition));
    $workflowTransition = $this->workflow->getTypePlugin()->getTransition($transition);
    \assert($workflowTransition instanceof EntityWorkflowTransition);
    $this->transition = $workflowTransition;

    $form['log'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Provide a log message'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): ContentEntityInterface {
    \assert($this->transition instanceof EntityWorkflowTransition);
    if ($this->transition->hasLog() && $this->transition->isLogRequired() && empty($form_state->getValue('log'))) {
      $form_state->setError($form['log'], $this->t('Log field is required'));
    }
    \assert($this->entity instanceof WorkspaceApprovalRequestInterface);
    return $this->entity;
  }

  /**
   * Checks access to the form.
   *
   * Suitable for a _custom_access route callback.
   * Based on the workspace_approval_request entity in the route.
   */
  public function access(AccountInterface $account, RouteMatchInterface $routeMatch): AccessResultInterface {
    $request = $routeMatch->getParameter('workspace_approval_request');
    if ($request instanceof WorkspaceApprovalRequestInterface) {
      return $request->access($this->operation, $account, TRUE);
    }
    return AccessResult::neutral();
  }

}
