<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\Core\Utility\Error;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;
use Drupal\workspaces\WorkspaceInterface;

/**
 * Provides a confirmation form to reject an approval request.
 */
final class WorkspaceApprovalRequestRejectForm extends WorkspaceApprovalRequestConfirmForm {

  /**
   * {@inheritdoc}
   */
  protected $operation = 'reject';

  /**
   * {@inheritdoc}
   */
  public function getQuestion(): TranslatableMarkup {
    return $this->t('Reject this %title?', [
      '%title' => $this->getBundleEntity()?->label(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription(): TranslatableMarkup {
    return $this->t("This action will prevent this content from transitioning to its next workflow state.");
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl(): Url {
    return Url::fromUserInput($this->getRedirectDestination()->get())->setAbsolute();
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    parent::submitForm($form, $form_state);
    \assert($this->entity instanceof WorkspaceApprovalRequestInterface);
    $this->entity->set('approved', FALSE);
    $this->entity->validate();
    $this->entity->save();
    $workspace = $this->entity->getSubject();
    \assert($workspace instanceof WorkspaceInterface);
    $form_state->setRedirectUrl(Url::fromRoute(
      'entity.workspace_approval_request.list.workspace',
      ['workspace' => $workspace->id()],
    ));

    $log = $form_state->getValue('log');

    try {
      \assert($this->workflow instanceof WorkflowInterface);
      \entity_workflow_transition($workspace, $this->workflow->id(), 'back_to_draft', $log);
      $this->messenger()->addMessage($this->t('The workflow state has been changed.'));
    }
    catch (\Exception $e) {
      Error::logException($this->logger('entity_workflow'), $e);
      $this->messenger()->addError($this->t('An error occurred while changing the workflow state. See logs for more details.'));
    }
  }

}
